"""Asynchronous Python client for WLED."""
import asyncio
import json
import socket
from typing import Any, Dict, Mapping, Optional, Tuple, Union

import aiohttp
import async_timeout
from yarl import URL

from .__version__ import __version__
from .exceptions import WLEDConnectionError, WLEDError
from .models import Device


class WLED:
    """Main class for handling connections with WLED."""

    device: Optional[Device] = None

    def __init__(
        self,
        host: str,
        base_path: str = "/json",
        password: str = None,
        port: int = 80,
        request_timeout: int = 8,
        session: aiohttp.client.ClientSession = None,
        tls: bool = False,
        username: str = None,
        verify_ssl: bool = True,
        user_agent: str = None,
    ) -> None:
        """Initialize connection with WLED."""
        self._session = session
        self._close_session = False

        self.base_path = base_path
        self.host = host
        self.password = password
        self.port = port
        self.socketaddr = None
        self.request_timeout = request_timeout
        self.tls = tls
        self.username = username
        self.verify_ssl = verify_ssl
        self.user_agent = user_agent

        if user_agent is None:
            self.user_agent = f"PythonWLED/{__version__}"

        if self.base_path[-1] != "/":
            self.base_path += "/"

    async def _request(
        self,
        uri: str = "",
        method: str = "GET",
        data: Optional[Any] = None,
        json_data: Optional[dict] = None,
        params: Optional[Mapping[str, str]] = None,
    ) -> Any:
        """Handle a request to a WLED device."""
        scheme = "https" if self.tls else "http"
        url = URL.build(
            scheme=scheme, host=self.host, port=self.port, path=self.base_path
        ).join(URL(uri))

        auth = None
        if self.username and self.password:
            auth = aiohttp.BasicAuth(self.username, self.password)

        headers = {
            "User-Agent": self.user_agent,
            "Accept": "application/json, text/plain, */*",
        }

        if self._session is None:
            self._session = aiohttp.ClientSession()
            self._close_session = True

        try:
            with async_timeout.timeout(self.request_timeout):
                response = await self._session.request(
                    method,
                    url,
                    auth=auth,
                    data=data,
                    json=json_data,
                    params=params,
                    headers=headers,
                    ssl=self.verify_ssl,
                )
        except asyncio.TimeoutError as exception:
            raise WLEDConnectionError(
                "Timeout occurred while connecting to WLED device."
            ) from exception
        except (aiohttp.ClientError, socket.gaierror) as exception:
            raise WLEDConnectionError(
                "Error occurred while communicating with WLED device."
            ) from exception

        content_type = response.headers.get("Content-Type", "")
        if (response.status // 100) in [4, 5]:
            contents = await response.read()
            response.close()

            if content_type == "application/json":
                raise WLEDError(response.status, json.loads(contents.decode("utf8")))
            raise WLEDError(response.status, {"message": contents.decode("utf8")})

        if "application/json" in content_type:
            return await response.json()

        return await response.text()

    async def update(self) -> Optional[Device]:
        """Get all information about the device in a single call."""
        try:
            data = await self._request()
            self.device = Device.from_dict(data)
        except WLEDError as exception:
            self.device = None
            raise exception

        return self.device

    async def light(
        self,
        brightness: Optional[int] = None,
        clones: Optional[int] = None,
        color_primary: Optional[
            Union[Tuple[int, int, int, int], Tuple[int, int, int]]
        ] = None,
        color_secondary: Optional[
            Union[Tuple[int, int, int, int], Tuple[int, int, int]]
        ] = None,
        color_tertiary: Optional[
            Union[Tuple[int, int, int, int], Tuple[int, int, int]]
        ] = None,
        effect: Optional[Union[int, str]] = None,
        intensity: Optional[int] = None,
        length: Optional[int] = None,
        on: Optional[bool] = None,
        palette: Optional[Union[int, str]] = None,
        reverse: Optional[bool] = None,
        segment_id: int = 0,
        selected: Optional[bool] = None,
        speed: Optional[int] = None,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        transition: Optional[int] = None,
    ) -> None:
        """Change state of a WLED Light segment."""
        if self.device is None:
            await self.update()

        if self.device is None:
            raise WLEDError("Unable to communicate with WLED to get the current state")

        device = self.device

        state = {
            "bri": brightness,
            "on": on,
        }

        segment = {
            "cln": clones,
            "fx": effect,
            "ix": intensity,
            "len": length,
            "pal": palette,
            "rev": reverse,
            "sel": selected,
            "start": start,
            "stop": stop,
            "sx": speed,
        }

        if effect is not None and isinstance(effect, str):
            segment["fx"] = next(
                (item.effect_id for item in self.device.effects if item.name == effect),
                None,
            )

        # Filter out not set values
        state = {k: v for k, v in state.items() if v is not None}
        segment = {k: v for k, v in segment.items() if v is not None}

        # Determine color set
        colors = []
        if color_primary is not None:
            colors.append(color_primary)
        elif color_secondary is not None or color_tertiary is not None:
            colors.append(device.state.segments[segment_id].color_primary)

        if color_secondary is not None:
            colors.append(color_secondary)
        elif color_tertiary is not None:
            colors.append(device.state.segments[segment_id].color_secondary)

        if color_tertiary is not None:
            colors.append(color_tertiary)

        if colors:
            segment["col"] = colors  # type: ignore

        if segment:
            segment["id"] = segment_id
            state["seg"] = [segment]  # type: ignore

        if transition is not None:
            state["transition"] = transition

        await self._request("state", method="POST", json_data=state)

        # Restore previous transition time
        if transition is not None:
            await self._request(
                "state",
                method="POST",
                json_data={"transition": device.state.transition},
            )

    async def transition(self, transition: int) -> None:
        """Set the default transition time for manual control."""
        await self._request(
            "state", method="POST", json_data={"transition": transition}
        )

    async def preset(self, preset: int) -> None:
        """Set a preset on a WLED device."""
        await self._request("state", method="POST", json_data={"ps": preset})

    async def playlist(self, playlist: int) -> None:
        """Set a running playlist on a WLED device."""
        await self._request("state", method="POST", json_data={"pl": playlist})

    async def sync(
        self, send: Optional[bool] = None, receive: Optional[bool] = None
    ) -> None:
        """Set the sync status of the WLED device."""
        sync = {"send": send, "recv": receive}
        sync = {k: v for k, v in sync.items() if v is not None}
        await self._request("state", method="POST", json_data={"udpn": sync})

    async def nightlight(
        self,
        duration: Optional[int] = None,
        fade: Optional[bool] = None,
        on: Optional[bool] = None,
        target_brightness: Optional[int] = None,
    ) -> None:
        """Control the nightlight function of a WLED device."""
        nightlight = {
            "dur": duration,
            "fade": fade,
            "on": on,
            "tbri": target_brightness,
        }

        # Filter out not set values
        nightlight = {k: v for k, v in nightlight.items() if v is not None}

        state: Dict[str, Any] = {"nl": nightlight}
        if on:
            state["on"] = True

        await self._request("state", method="POST", json_data=state)

    async def close(self) -> None:
        """Close open client session."""
        if self._session and self._close_session:
            await self._session.close()

    async def __aenter__(self) -> "WLED":
        """Async enter."""
        return self

    async def __aexit__(self, *exc_info) -> None:
        """Async exit."""
        await self.close()



"""Asynchronous Python client for WLED."""

from .models import (  # noqa
    Device,
    Effect,
    Info,
    Leds,
    Nightlight,
    Palette,
    Segment,
    State,
    Sync,
)
from .wled import WLED, WLEDConnectionError, WLEDError  # noqa



"""Models for WLED."""

from typing import List, Tuple, Union

import attr


@attr.s(auto_attribs=True, frozen=True)
class Nightlight:
    """Object holding nightlight state in WLED."""

    duration: int
    fade: bool
    on: bool
    target_brightness: int

    @staticmethod
    def from_dict(data):
        """Return Nightlight object from WLED API response."""
        nightlight = data.get("nl", {})
        return Nightlight(
            duration=nightlight.get("dur", 1),
            fade=nightlight.get("fade", False),
            on=nightlight.get("on", False),
            target_brightness=nightlight.get("tbri"),
        )


@attr.s(auto_attribs=True, frozen=True)
class Sync:
    """Object holding sync state in WLED."""

    send: bool
    receive: bool

    @staticmethod
    def from_dict(data):
        """Return Sync object from WLED API response."""
        sync = data.get("udpn", {})
        return Sync(send=sync.get("send", False), receive=sync.get("recv", False))


@attr.s(auto_attribs=True, frozen=True)
class Effect:
    """Object holding an effect in WLED."""

    effect_id: int
    name: str


@attr.s(auto_attribs=True, frozen=True)
class Palette:
    """Object holding an palette in WLED."""

    palette_id: int
    name: str


@attr.s(auto_attribs=True, frozen=True)
class Segment:
    """Object holding segment state in WLED."""

    segment_id: int
    start: int
    stop: int
    length: int
    color_primary: Union[Tuple[int, int, int, int], Tuple[int, int, int]]
    color_secondary: Union[Tuple[int, int, int, int], Tuple[int, int, int]]
    color_tertiary: Union[Tuple[int, int, int, int], Tuple[int, int, int]]
    effect: Effect
    speed: int
    intensity: int
    palette: Palette
    selected: bool
    reverse: bool
    clones: int

    @staticmethod
    def from_dict(
        segment_id: int, data: dict, effects: List[Effect], palettes: List[Palette]
    ) -> "Segment":
        """Return Segment object from WLED API response."""
        start = data.get("start", 0)
        stop = data.get("stop", 0)
        length = data.get("len", (stop - start))

        colors = data.get("col", [])
        primary_color, secondary_color, tertiary_color = (0, 0, 0)
        try:
            primary_color = tuple(colors.pop(0))  # type: ignore
            secondary_color = tuple(colors.pop(0))  # type: ignore
            tertiary_color = tuple(colors.pop(0))  # type: ignore
        except IndexError:
            pass

        effect = next(
            (item for item in effects if item.effect_id == data.get("fx", 0)),
            Effect(effect_id=0, name="Unknown"),
        )
        palette = next(
            (item for item in palettes if item.palette_id == data.get("pal", 0)),
            Palette(palette_id=0, name="Unknown"),
        )

        return Segment(
            segment_id=segment_id,
            start=start,
            stop=stop,
            length=length,
            color_primary=primary_color,  # type: ignore
            color_secondary=secondary_color,  # type: ignore
            color_tertiary=tertiary_color,  # type: ignore
            effect=effect,
            speed=data.get("sx", 0),
            intensity=data.get("ix", 0),
            palette=palette,
            selected=data.get("sel", False),
            reverse=data.get("reverse", False),
            clones=data.get("cln", -1),
        )


@attr.s(auto_attribs=True, frozen=True)
class Leds:
    """Object holding leds info from WLED."""

    count: int
    rgbw: bool
    pin: int
    power: int
    max_power: int
    max_segments: int

    @staticmethod
    def from_dict(data: dict):
        """Return Leds object from WLED API response."""
        leds = data.get("leds", {})
        return Leds(
            count=leds.get("count", 0),
            rgbw=leds.get("rgbw", False),
            pin=leds.get("pin", 0),
            power=leds.get("pwr", 0),
            max_power=leds.get("maxpwr", 0),
            max_segments=leds.get("maxseg", 0),
        )


@attr.s(auto_attribs=True, frozen=True)
class Info:
    """Object holding information from WLED."""

    architecture: str
    arduino_core_version: str
    brand: str
    build_type: str
    effect_count: int
    free_heap: int
    leds: Leds
    live: bool
    mac_address: str
    name: str
    pallet_count: int
    product: str
    udp_port: int
    uptime: int
    version_id: str
    version: str

    @staticmethod
    def from_dict(data: dict):
        """Return Info object from WLED API response."""
        return Info(
            architecture=data.get("arch", "Unknown"),
            arduino_core_version=data.get("core", "Unknown").replace("_", "."),
            brand=data.get("brand", "WLED"),
            build_type=data.get("btype", "Unknown"),
            effect_count=data.get("fxcount", 0),
            free_heap=data.get("freeheap", 0),
            leds=Leds.from_dict(data),
            live=data.get("live", False),
            mac_address=data.get("mac", ""),
            name=data.get("name", "WLED Light"),
            pallet_count=data.get("palcount", 0),
            product=data.get("product", "DIY Light"),
            udp_port=data.get("udpport", 0),
            uptime=data.get("uptime", 0),
            version_id=data.get("vid", "Unknown"),
            version=data.get("ver", "Unknown"),
        )


@attr.s(auto_attribs=True, frozen=True)
class State:
    """Object holding the state of WLED."""

    nightlight: Nightlight
    sync: Sync
    segments: List[Segment]
    on: bool
    brightness: int
    transition: int
    preset: int
    playlist: int

    @property
    def playlist_active(self):
        """Return if a playlist is currently active."""
        return self.playlist == -1

    @property
    def preset_active(self):
        """Return if a preset is currently active."""
        return self.preset == -1

    @staticmethod
    def from_dict(data, effects: List[Effect], palettes: List[Palette]):
        """Return State object from WLED API response."""
        segments = []
        for segment_id, segment in enumerate(data.get("seg", [])):
            segments.append(Segment.from_dict(segment_id, segment, effects, palettes))

        return State(
            nightlight=Nightlight.from_dict(data),
            sync=Sync.from_dict(data),
            segments=segments,
            on=data.get("on", False),
            brightness=data.get("bri", 1),
            transition=data.get("transition", 0),
            preset=data.get("ps", -1),
            playlist=data.get("pl", -1),
        )


@attr.s(auto_attribs=True, frozen=True)
class Device:
    """Object holding all information of WLED."""

    effects: List[Effect]
    info: Info
    palettes: List[Palette]
    state: State

    @staticmethod
    def from_dict(data):
        """Return Device object from WLED API response."""
        effects = []
        for effect_id, effect in enumerate(data.get("effects", {})):
            effects.append(Effect(effect_id=effect_id, name=effect))
        effects.sort(key=lambda x: x.name)

        palettes = []
        for palette_id, palette in enumerate(data.get("palettes", {})):
            palettes.append(Palette(palette_id=palette_id, name=palette))
        palettes.sort(key=lambda x: x.name)

        return Device(
            effects=effects,
            info=Info.from_dict(data.get("info", {})),
            palettes=palettes,
            state=State.from_dict(data.get("state", {}), effects, palettes),
        )
